/* Foefel.java -- various helper functions for analysis of GUI components as well as static string properties.
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package evilwan;

import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.JPanel;
import java.util.*;
import java.io.*;
import javax.swing.event.*;
import java.lang.reflect.*;

public class Foefel {
    // {{ Class members }}

    // {{ Instance members }}
    
    // {{ Class methods }}

    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("Foefel -- " + s);
    }
    //
    // Utility methods for dumping various types of static properties.
    //
    /**
     * Format a string using more or less standard Java syntax for special characters.
     */
    private static String s2s(String s) {
	StringBuilder sb = new StringBuilder();
	for(int i = 0; i < s.length(); ++i) {
	    char c = s.charAt(i);
	    if((c == '\\') || (c == '\'') || (c == '\"')){
		sb.append('\\');
		sb.append(c);
	    } else if(c == '\n') {
		sb.append('\\');
		sb.append('n');
	    } else if(c == '\r') {
		sb.append('\\');
		sb.append('r');
	    } else if(c == '\f') {
		sb.append('\\');
		sb.append('f');
	    } else if(c == '\t') {
		sb.append('\\');
		sb.append('t');
	    } else if(((int)c < 0x20) || ((int)c > 0x7E)) {
		sb.append(String.format("\\u%04x", (int)c));
	    } else {
		sb.append(c);
	    }
	}
	return sb.toString();
    }
    /**
     * Format a byte[] showing both hex values and character representations (if applicable)
     */
    private static String b2s(byte[] bar) {
	StringBuilder sb = new StringBuilder();
	StringBuilder sa = new StringBuilder();
	for(int i = 0; i < bar.length; ++i) {
	    if((i % 16) == 0) {
		if(i > 0) {
		    sb.append("\t// ");
		    sb.append(sa.toString());
		    sb.append('\n');
		    sa.delete(0, sa.length());
		}
		sb.append('\t');
	    }
	    int byt = ((int)bar[i]) & 0xff;
	    sb.append(String.format("0x%02x, ", byt));
	    if((byt >= 0x20) && (byt < 0x7f)) {
		sa.append((char) byt);
	    } else {
		sa.append('.');
	    }
	}
	if(sa.length() > 0) {
	    sb.append("\t// ");
	    sb.append(sa.toString());
	}
	return sb.toString();
    }
    /**
     * Format a char[] using more or less standard Java syntax for special characters.
     */
    private static String c2s(char[] s) {
	StringBuilder sb = new StringBuilder();
	for(int i = 0; i < s.length; ++i) {
	    char c = s[i];
	    if((c == '\\') || (c == '\'') || (c == '\"')){
		sb.append('\\');
		sb.append(c);
	    } else if(c == '\n') {
		sb.append('\\');
		sb.append('n');
	    } else if(c == '\r') {
		sb.append('\\');
		sb.append('r');
	    } else if(c == '\f') {
		sb.append('\\');
		sb.append('f');
	    } else if(c == '\t') {
		sb.append('\\');
		sb.append('t');
	    } else if(((int)c < 0x20) || ((int)c > 0x7E)) {
		sb.append(String.format("\\u%04x", (int)c));
	    } else {
		sb.append(c);
	    }
	}
	return sb.toString();
    }

    private static String dumpClassHierarchy(Class c, String ind) {
	StringBuilder sb = new StringBuilder();
	sb.append(ind);
	sb.append("Class: ");
	sb.append(c.getName());

	Class mom = c.getSuperclass();
	if(mom == null) {
	    return sb.toString();
	}
	sb.append(" extends ");
	sb.append(mom.getName());
	while(!mom.getName().startsWith("java")) {
	    mom = mom.getSuperclass();
	    sb.append(" -> ");
	    sb.append(mom.getName());
	}
	
	//Class[] interfaces = c.getInterfaces();
	//if(interfaces.length > 0) {
	//    sb.append(" implements ");
	//    for(int i = 0; i < interfaces.length; ++i) {
	//	if(i > 0) {
	//	    sb.append(", ");
	//	}
	//	sb.append(interfaces[i].getName());
	//    }
	//}
	//if(!mom.getName().equals("java.lang.Object")) {
	if(!mom.getName().startsWith("java")) {
	    sb.append('\n');
	    sb.append(dumpClassHierarchy(mom, ind + "    "));
	}
	return sb.toString();
    }
    //
    // Various GUI component analysis methods
    //
    private static Field findListProperty(Object o) {
	Field fld = null;
	try {
	    String s = ObjectDumper.dumpTableModelObject(o);
	    say("####### findListProperty:\n" + s);
	    Class c = o.getClass();
	    Field[] flds = c.getDeclaredFields();
	    say("findListProperty() -- checking " + flds.length + " fields in class " + c.getName());
	    for(int i = 0; i < flds.length; ++i) {
		flds[i].setAccessible(true);	// Make sure we can access the field
		say("findListProperty() -- field " + flds[i].getName() + " is a " + flds[i].getType().getName());
		say("findListProperty() -- field " + flds[i].getName() + " class hierarchy: " + dumpClassHierarchy(flds[i].getType(), ""));
		if(flds[i].getType().equals(ArrayList.class)) {
		    if(fld != null) {
			throw new Exception("Instance of " + o.getClass().getName() +
					    " contains multiple List properties: found " +
					    fld.toGenericString() + " and " +
					    flds[i].toGenericString());
		    }
		    fld = flds[i];
		}
	    }
	} catch(Throwable t) {
	    System.out.println("## caught: " + t);
	    t.printStackTrace();
	    System.exit(1);
	}
	return fld;
    }

    private static JTable analyzeJTable(JTable tab) {
	TableModel model = tab.getModel();
	int n = model.getColumnCount();
	say("analyzeJTable() -- model.getColumnCount() = " + n);
	for(int i = 0; i < n; ++i) {
	    String colnam = model.getColumnName(i);
	    say("analyzeJTable() -- colnam = " + colnam);
	    //
	    // Hack: in current release the only table with a column "Direction" is
	    // the JTable for websocket traffic
	    //
	    if(colnam.equals("Direction")) {
		//
		// Gotcha!
		//
		//_websockets_table = tab;
		say("analyzeJTable() -- gotcha!");
		return tab;
	    }
	}
	return analyzeKids(tab);
    }

    private static JTable analyzeKids(Container box) {
	Component[] parts = box.getComponents();
	JTable jt = null;
	say("analyzeKids() -- parts.length = " + parts.length);
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof JTable) {
		say("analyzeKids() --  found JTable!");
		jt = analyzeJTable((JTable) parts[i]);
	    } else if(parts[i] instanceof Container) {
		jt = analyzeKids((Container) parts[i]);
	    }
	    if(jt != null) {
		return jt;
	    }
	}
	return jt;
    }
    /**
     * Recursively look for a JTable with specified column name.
     */
    private static JTable lookForWaldo(Container box, String waldo) {
	JTable jt = null;
	//
	// Check if we are the one
	//
	if(box instanceof JTable) {
	    JTable jtbox = (JTable) box;
	    int n = jtbox.getColumnCount();
	    for(int i = 0; i < n; ++i) {
		String colnam = jtbox.getColumnName(i);
		if(colnam.equals(waldo)) {
		    return jtbox;
		}
	    }
	}
	//
	// Ok, not the one: descend one level and try again with each kid
	//
	Component[] parts = box.getComponents();
	for(int i = 0; i < parts.length; ++i) {
	    if(parts[i] instanceof Container) {
		jt = lookForWaldo((Container) parts[i], waldo);
		if(jt != null) {
		    return jt;
		}
	    }
	}
	return jt;
    }
    
    /**
     * Find object that contains the actual websocket messages
     */
    public static Object findWSTable() throws Exception {
	JTable jt = null;
	//
	// Loop over all top-level frames and scan each recursively until we
	// find a JTable that has a column named "Direction" (because that is the
	// JTable that contains the websocket info: there is only one JTable with
	// that column in the entire interface)
	//
	Frame[] frames = Frame.getFrames();
	for(int i = 0; i < frames.length; ++i) {
	    //
	    // Now analyze entire frame component tree until we
	    // find the JTable component containing the websocket packet list
	    //
	    // Only one JTable in entire GUI with a column named "Direction"...
	    //
	    jt = lookForWaldo(frames[i], "Direction");
	    //say("findWSTable() -- JTable containing websocket packets: " +
	    //((jt == null)?"<null>":"found"));
	    if(jt != null) {
		//
		// Gotcha!
		//
		break;
	    }
	}
	//
	// Now find remaining pieces of information from Burp environment: we intend
	// to find some sort of List-type property in the TableModel linked to the JTable
	// object that we found in the GUI. This can be done by simply dumping the full class name
	// of the TableModel of the JTable object that we have in "_websockets_table" (for Burp 1.6
	// this results in "burp.dzb" but this is most likely different for each build of Burp)
	//
	// Using javap on "burp.dzb" shows only one List-type property (an ArrayList) and examining the
	// bytecode for method "burp.dzb.getRowCount()" confirms that the "size()" of that ArrayList property
	// is returned as rowcount in the TableModel. Again, in Burp 1.6, the name of that property is "t" (that
	// name is not required in the analysis code below, just that it's an ArrayList property)
	//
	// The assumption is that whatever is stored in that ArrayList, is some sort of encapsulation of all
	// data relevant for each websocket packet (timestamp, direction, original message data, modified message
	// data etc.) So, next we need to get hold of that ArrayList and then we need to find a way to
	// interpret the contents of each of its elements.
	//
	// Note that, thanks to reflection, we do have access to the private ArrayList property after
	// using the Field.setAccessible(true) call :-)
	//
	TableModel model = jt.getModel();
	Field f = findListProperty(model);
	Object ooo = f.get(model);
	return ooo;	// return actual ArrayList property
    }
    /**
     * Return a list of static class properties that are either strings,
     * character arrays or byte arrays.
     */
    public static Vector<String> getStaticStrings(Class c) throws Exception {
	String clsnam = c.getName();
	int jj = clsnam.lastIndexOf(".");
	String shortname = clsnam.substring(jj + 1);
	Vector<String> v = new Vector<String>();
	Field[]	flds = c.getDeclaredFields();
	for(int i=0; i < flds.length; ++i) {
	    flds[i].setAccessible(true);	// Make sure we can access the field
	    String fldnam = flds[i].getName();
	    if("[Ljava.lang.String;".equals(flds[i].getType().getName())) {
		//
		// String array
		//
		try {
		    Object o = flds[i].get(null);
		    for(int j = 0; j < Array.getLength(o); ++j) {
			v.add(shortname + "." + fldnam + "[" + j + "]=\"" + s2s((String)Array.get(flds[i].get(null), j)) + "\";");
		    }
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    } else if ("java.lang.String".equals(flds[i].getType().getName())) {
		//
		// Scalar String
		//
		try {
		    v.add(shortname + "." + fldnam + "=\"" + s2s((String)flds[i].get(null)) + "\";");
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    } else if ("[B".equals(flds[i].getType().getName())) {
		//
		// byte array
		//
		try {
		    v.add(shortname + "." + fldnam + "={\n" + b2s((byte[])flds[i].get(null)) + "\n};");
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    } else if ("[C".equals(flds[i].getType().getName())) {
		//
		// char array
		//
		try {
		    v.add(shortname + "." + fldnam + "={\n" + c2s((char[])flds[i].get(null)) + "\n};");
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    }
	}
	return v;
    }

    public static void dumpStatics(Class c) {
	try {
	    Vector<String> strings = getStaticStrings(c);
	    for(Enumeration<String> e = strings.elements(); e.hasMoreElements(); ) {
		say(e.nextElement());
	    }
	} catch (Exception ex) {
	    say("caught: " + ex);
	    ex.printStackTrace();
	}
    }

    // {{ Constructors }}
    
    
    // {{ Instance methods }}

}
