/* Geronimo.java - Utility for dumping static class string and byte array properties.
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package evilwan;

import java.lang.reflect.*;
import java.util.*;
import java.io.*;

/**
 * Utility for dumping static class string and byte array properties.
 *
 * The main reason for this class is that the Burp class files are the output
 * of an obfuscation tool that, amongst other things, seems to mangle all string
 * literals. The algorithm that is used by that tool is to replace all string literal
 * references by references to elements in a private static string array. Those arrays
 * are initialized with an obfuscated form of the plain text. Additional code is
 * added to the class to unmangle those strings when the class is loaded: static
 * class initialization code is reponsible for that demangling phase.
 * <p>
 * The effect is that when the class is loaded by the classloader, the static code
 * for string demangling has already been executed so that the original string literals
 * are stored in private static string arrays. The code below is intended to be
 * used to read those strings after the classes have been loaded. Java reflection allows
 * go find those private static fields, and also to read their values (after a call to
 * explicitly grant access to those fields) The end result is that we don't care about
 * the mangling/demangling logic, beause we can go directly to the end result.
 */
public class Geronimo {
    // {{ Class members }}
    
    //
    // For debugging purposes (there is a "main()" at the botto of this class): dump
    // this class' static string properties
    //
    private static String S1 = "this is\ta \'private\"\nstatic String";
    private static String[] S2 = {"this is a string array", "with two elements"};
    private static byte[] bar = {0x41, 0x42, 0x43, 0x44};

    // {{ Instance members }}
    
    // {{ Class methods }}
    
    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("Geronimo -- " + s);
    }

    /**
     * Escape special characters in a quoted string.
     */
    private static String s2s(String s) {
	StringBuilder sb = new StringBuilder();
	for(int i = 0; i < s.length(); ++i) {
	    char c = s.charAt(i);
	    if((c == '\\') || (c == '\'') || (c == '\"')){
		sb.append('\\');
		sb.append(c);
	    } else if(c == '\n') {
		sb.append('\\');
		sb.append('n');
	    } else if(c == '\r') {
		sb.append('\\');
		sb.append('r');
	    } else if(c == '\f') {
		sb.append('\\');
		sb.append('f');
	    } else if(c == '\t') {
		sb.append('\\');
		sb.append('t');
	    } else if(((int)c < 0x20) || ((int)c > 0x7E)) {
		sb.append(String.format("\\u%04x", (int)c));
	    } else {
		sb.append(c);
	    }
	}
	return sb.toString();
    }
    private static String b2s(byte[] bar) {
	StringBuilder sb = new StringBuilder();
	StringBuilder sa = new StringBuilder();
	for(int i = 0; i < bar.length; ++i) {
	    if((i % 16) == 0) {
		if(i > 0) {
		    sb.append("\t// ");
		    sb.append(sa.toString());
		    sb.append('\n');
		    sa.delete(0, sa.length());
		}
		sb.append('\t');
	    }
	    int byt = ((int)bar[i]) & 0xff;
	    sb.append(String.format("0x%02x, ", byt));
	    if((byt >= 0x20) && (byt < 0x7f)) {
		sa.append((char) byt);
	    } else {
		sa.append('.');
	    }
	}
	if(sa.length() > 0) {
	    sb.append("\t// ");
	    sb.append(sa.toString());
	}
	return sb.toString();
    }
    private static String c2s(char[] s) {
	StringBuilder sb = new StringBuilder();
	for(int i = 0; i < s.length; ++i) {
	    char c = s[i];
	    if((c == '\\') || (c == '\'') || (c == '\"')){
		sb.append('\\');
		sb.append(c);
	    } else if(c == '\n') {
		sb.append('\\');
		sb.append('n');
	    } else if(c == '\r') {
		sb.append('\\');
		sb.append('r');
	    } else if(c == '\f') {
		sb.append('\\');
		sb.append('f');
	    } else if(c == '\t') {
		sb.append('\\');
		sb.append('t');
	    } else if(((int)c < 0x20) || ((int)c > 0x7E)) {
		sb.append(String.format("\\u%04x", (int)c));
	    } else {
		sb.append(c);
	    }
	}
	return sb.toString();
    }

    public static Vector<String> getStaticStrings(String clsnam) throws Exception {
	Vector<String> v = new Vector<String>();
	Class	c = Class.forName(clsnam);
	Field[]	flds = c.getDeclaredFields();
	for(int i=0; i < flds.length; ++i) {
	    flds[i].setAccessible(true);	// Make sure we can access the field
	    String fldnam = flds[i].getName();
	    if("[Ljava.lang.String;".equals(flds[i].getType().getName())) {
		//
		// String array
		//
		try {
		    Object o = flds[i].get(null);
		    for(int j = 0; j < Array.getLength(o); ++j) {
			v.add(fldnam + "[" + j + "]=\"" + s2s((String)Array.get(flds[i].get(null), j)) + "\";");
		    }
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    } else if ("java.lang.String".equals(flds[i].getType().getName())) {
		//
		// Scalar String
		//
		try {
		    v.add(fldnam + "=\"" + s2s((String)flds[i].get(null)) + "\";");
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    } else if ("[B".equals(flds[i].getType().getName())) {
		//
		// byte array
		//
		try {
		    v.add(fldnam + "={\n" + b2s((byte[])flds[i].get(null)) + "\n};");
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    } else if ("[C".equals(flds[i].getType().getName())) {
		//
		// char array
		//
		try {
		    v.add(fldnam + "={\n" + c2s((char[])flds[i].get(null)) + "\n};");
		} catch(NullPointerException npe) {
		    String msg = clsnam + "." + fldnam + " not available: modifiers = " + flds[i].getModifiers() +
			" (" + Modifier.toString(flds[i].getModifiers()) + ")";
		    say(msg);
		    v.add(msg);
		}
	    }
	}
	return v;
    }

    /**
     * Dumps all static string declarations in specified class in a log file.
     *
     * @param filnam name of class file
     * @param outsuffix suffix, including leading dot, to use for output dump file
     */
    public static void dumpClassStrings(String filnam, String prefix, String outsuffix) throws Exception {
	int  i = filnam.lastIndexOf(".");
	if(i < 0) {
	    throw new Exception("No period found in filename: " + filnam);
	}
	String clsnam = filnam.substring(prefix.length(), i).replaceAll("/", ".");
	String outfilnam = filnam.substring(0, i) + outsuffix;
	say("dumping class " + clsnam + " ...");
	Vector<String> strings = getStaticStrings(clsnam);
	FileWriter o = new FileWriter(outfilnam);
	o.write("// " + clsnam + "\n");
	for(Enumeration<String> e = strings.elements(); e.hasMoreElements(); ) {
	    o.write(e.nextElement());
	    o.write("\n");
	}
	o.close();
    }

    /**
     * Dump static class string properties of all classes listed in <code>lstfil</code>
     *
     * @param lstfil file containing the filenames of all classes to dump, one line per classfile
     * @param outsuffix suffix, including leading dot, to use for output dump files
     */
    public static void dumpClassesFromFile(String lstfil, String prefix, String outsuffix) throws Exception {
	say("~ dumpClassesFromFile -- using list file: " + lstfil);
	FileReader fr = new FileReader(lstfil);
	BufferedReader is = new BufferedReader(fr);
	String line;
	while((line = is.readLine()) != null) {
	    line = line.replaceAll("#.*$", "").trim();
	    if(line.length() > 0) {
		try {
		    dumpClassStrings(line, prefix, outsuffix);
		} catch(Throwable t) {
		    say("caught: " + t);
		    t.printStackTrace();
		}
	    }
	}
	is.close();
	fr.close();
    }

    public static void dumpClassFile(String clsnam, String clspad) throws Exception {
	Class	c = Class.forName(clsnam);
	ClassLoader ldr = c.getClassLoader();
	InputStream is = ldr.getResourceAsStream(clspad);
	if(is == null) {
	    say("dumpClassFile() -- failed to get resource " + clspad);
	} else {
	    say("dumpClassFile() -- got resource " + clspad);
	    is.close();
	}
    }

    public static void main(String[] args) {
	try {
	    if((args.length > 2) && (args[0].equals("-D"))) {
		dumpClassFile(args[1], args[2]);
	    } else {
		int i = 0;
		while(i < args.length) {
		    dumpClassesFromFile(args[i++], args[i++], ".dump");
		    ++i;
		}
	    }
	} catch(Throwable t) {
	    say("caught: " + t);
	    t.printStackTrace();
	}
    }
    // {{ Constructors }}
    
    /**
     * Default public constructor.
     */
    public Geronimo() {
    }
    
    // {{ Instance methods }}
    
}
