/* BurpExtender.java - Burp extender main class
 *
 * Two options are added to BurpSuitePro:
 *
 * 1. save intercepted websocket messages in an XML file
 * 2. use an automatic filter to process the data of every websocket message
 *
 * Automated modification of websockets messages requires the implementation of
 * a Java class that implements the interface evilwan.WSFIlter That implementation class
 * should have a constructor that is to be called without arguments.
 *
 * Note: this implementation of a Burp extension only works with Burp Professional version 1.7.15
 *
 * In order to implement this extension, a number of educated guesses were used (such as the assumption
 * that the table of intercepted websocket messases is based on a class that encapsulates all relevant
 * data for each websocket message), as well as dumping the entire Burp GUI compoment tree and, in a
 * couple of situations, even decompiling obfuscated class files with "procyon". Additional information was
 * found using Java reflection.
 *
 * In brief: it is the expectation (hope) of the author that the functionality covered by this
 * extension will be provided by the next release of Burp, but in the mean time this extension does
 * allow to save websocket trafic in an XML file, as well as providing automatic modifications of
 * websocket messages by means of a custom filter.
 *
 * This edition comes with most of the debugging/analysis tools that were used to reverse those parts
 * of Burp that are required to implement the additional functionality. The idea behind that is that
 * when a new version of Burp is released, the user has the tools and knowledge to modify the code
 * in this extension to match the new version. A working knowledge of Java Swing and Java reflection
 * is required, though.
 *
 * Enjoy.
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package burp;

import evilwan.*;
import java.awt.*;

public class BurpExtender implements IBurpExtender, ITab {
    // {{ Class members }}

    /**
     * Tab title
     */
    private final static String TAB_CAPTION = "WS Stuff";
    /**
     * Identification and version string shown in the Burp overview of loaded extensions
     */
    private final static String EXT_NAME = "WebSockets Stuff v1.7.15";

    // {{ Instance members }}
    
    /**
     * Burp callbacks
     */
    private IBurpExtenderCallbacks	_cb = null;
    /**
     * Additional pane used as GUI for this extension
     */
    private Component			_tabpane = null;

    // {{ Class methods }}
    
    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("BurpExtender -- " + s);
    }
    // {{ Constructors }}
    
    // {{ Instance methods }}

    //
    // IBurpExtender interface
    //
    /**
     * Main entry point for Burp extensions
     */
    public void registerExtenderCallbacks(IBurpExtenderCallbacks cb) {
	_cb = cb;
	_cb.setExtensionName(EXT_NAME);
	_tabpane = new WSGui();
	_cb.addSuiteTab(this);
	//
	// Reversing aids: remaining code in this method can safely be removed
	// in operational version, but can be of use for modifying the code for a
	// new version of BurpSuite.
	//
	String[] args = _cb.getCommandLineArguments();
	int i = 0;
	while(i < args.length) {
	    if(args[i].equals("-strings")) {
		//
		// Read list of class files (one line per class file) and for
		// each of the classes dump the static strings and byte arrays
		// found in those classes.
		//
		// The obfuscator used on BurpSuite stores those string and byte
		// array contents in a mangled form but at class load time those
		// values are demangled. We don't care about the actual mangling
		// algorithm, we just wait until the class loader finishes the
		// loading after which the values will be available in plain text
		// format.
		//
		// Expected format of input lines is like so:
		//
		//     unpack/burp/abc.class
		//
		// The "-strings" parameter is followed by 2 more arguments:
		//
		// 1. the filename of the list of classes to process
		// 2. the prefx string to remove from each entry before using it as
		//    an external class refereence (in the example above that would
		//    be "unpack/")
		//
		// For each of the specified class files there will be a
		// corresponding ".dump" created.
		//
		// The .dump files can easily be merged with the output of Procyon
		// to produce more readable Java code.
		//
		// E.g.:
		//
		// $ cd unpack/burp
		// $ java -jar ../../procyon.jar -eta abc.class > abc.procy
		// $ ../../merge-strings.pl abc
		// $ less abc.java
		//
		try {
		    Geronimo.dumpClassesFromFile(args[++i], args[++i], ".dump");
		} catch(Throwable t) {
		    say("caught: " + t);
		    t.printStackTrace();
		}
	    } else if(args[i].equals("-dumpgui")) {
		//
		// Dump the entire GUI structure of the Burp application interface.
		//
		// Works best with standard output redirected to a log file. E.g.:
		//
		// $ java -classpath burp.jar:. burp.StartBurp -dumpgui > gui.log
		//
		// Main point of interest in that output is the JTable with a
		// column named "Direction": so far there is only one of those in the
		// entire API :-)
		//
		try {
		    Goewie.dumpBurpComponents();
		} catch(Throwable t) {
		    say("caught: " + t);
		    t.printStackTrace();
		}
	    } else if(args[i].equals("-analyze")) {
		//
		// Used for more in-depth analysis of the run-time object structure
		// of burp specific objects.
		//
		// This option will do nothing except adding a special tab to the
		// interface with only one button. The plan is to wait until a couple
		// of websocket messages have been intercepted and then click on the
		// modedst "Go" button in that tab.
		//
		// Note that the code in ObjectDumper might have to be commented out
		// (partially) at first until the new names of the Burp specific
		// classes are discovered. Start by dumping the TableModel for the
		// JTable and see which of its properties looks like it could be a
		// list, array or some other collection. Then add code to dump the
		// first element in that list/array/collection. Use that output to
		// make educated guesses about which methods to use on the "element"
		// objects to extract the index, timestamp, original message data
		// etc.
		//
		// Beware: the obfuscator that is used on BurpSuite has an
		// interesting trick up its sleave. Normally the "signature" of a
		// Java method should be unique (the return value of a method is not
		// part of that signature) However, at Java VM level the return
		// type *is* part of the method definitions. The obfuscator will
		// generate, for instance, methods like this:
		//
		// public boolean a() { return some_boolean; }
		// public int a() { return some_int; }
		// public long a() { return some_long; }
		// public String a() { return some_string; }
		// public burp.abc a() { return some_burp_abc_instance; }
		//
		// All methods have the same signature for the Java compiler, but not
		// for the Java VM. One result is that the above Java fragment will
		// not compile (but Procyon *will* return all the methods) Another
		// result is that it is not possible with Java reflection to query
		// for method "a()": the call will return with a method, but not
		// necessarily the one you need. Fortunately it is possible to get
		// the complete list of methods using Java reflection: just ask for
		// *all* methods instead of on name/signature.
		//
		try {
		    _tabpane = new Goewie();
		    _cb.addSuiteTab(this);
		} catch(Throwable t) {
		    say("caught: " + t);
		    t.printStackTrace();
		}
	    } else if(args[i].equals("-XXX")) {
		//
		// Activate debugging features in WSPacket
		//
		WSPacket.setDebug(true);
	    }
	}
    }
    
    //
    // ITab interface
    //
    public String getTabCaption() {
	return TAB_CAPTION;
    }
    public Component getUiComponent() {
	return _tabpane;
    }
}
