/* WSPacket -- Wrapper around data kept by Burp for each websocket message.
 *
 * This class is based on the assumption that Burp encapsulates all properties
 * for one websocket message (time, direction, data...) in one class and that
 * the TableModel of the JTable that shows an overview of all intercepted
 * websocket messages, has some sort of array/list/collection type property
 * containing one of those websocket message encapsulation objects per entry
 * (table row)
 *
 * The general plan for adapting this class to each new version of BurpSuite
 * goes something like this:
 *
 * 1. at Burp run-time, scan all application top frames and all of their children
 *    recursively until a JTable object is found that has a colunm name "Direction"
 *
 * 2. get the TableModel object for that JTable object: note the class name for
 *    that object
 *
 * 3. use Procyon to decompile the class of which the TableModel is an instance
 *
 * 4. use the information of that decompiled version to find the names of all
 *    instance properties and use ObjectDumper to dump all those properties of
 *    the actual TableModel object (note that this step requires that there are
 *    at least some websocket message objects in the table: that is why there
 *    is a debug feature made available that allows manual triggering of this
 *    analysis step) The code in ObjectDumper will attempt to print additional
 *    information on the first element of array/list/collection type of
 *    properties)
 *
 * 5. if one of the properties looks like a good candidate for the actual list
 *    of stored websocket messages, (re)activate the code in ObjectDumper to
 *    also dump the properties of the first element in that assumed messages
 *    table
 *
 * 6. use Procyon to decompile the Burp class of which that first element is
 *    an instance
 *
 * 7. based on the data dumped in the previous steps, try to find the accessor
 *    methods on the row element objects for:
 *
 *    - index (int)
 *    - URL (a Burp specific class)
 *    - message data encapsulation (a Burp specific class): there are 3
 *      properties with this type (original message, automodified message
 *      and edited message)
 *    - timestamp (long)
 *    - direction (boolean)
 *    - highlight code (byte)
 *    - comment, optional (can be null, String type)
 *    - local port number (int)
 *
 *    There are more properties, but the above set is used by the code below.
 *
 *    Apparently the constructor for the array/list element objects has the
 *    following arguments (and in that order):
 *
 *    - int local-port
 *    - URL encapsulation
 *    - original message data
 *    - auto modified message data
 *    - highlight code
 *    - message data type enumeration code
 *    - timestamp
 *
 *    This set of constructor arguments has not been changed for the last
 *    versions of BurpSuite.
 *
 * The complete list of accessor methods on the websocket message encapsulation
 * objects is:
 *
 * boolean c()	-> direction
 * int c()	-> index
 * burp.qpd c()	-> URL encapsulation
 * String a()	-> comment (optional, can return null)
 * int a()	-> local port
 * byte a()	-> highlight
 * burp.jme f()	-> original message
 * long f()	-> timestamp
 * burp.jme d()	-> edited message
 * burp.jme e()	-> automodified message
 * int g()	-> ???
 * burp.vld b()	-> ???
 * burp.wog b()	-> returns message data type enumeration
 *
 * Copyright (c) 2017 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package evilwan;

import java.util.*;
import java.lang.reflect.*;

public class WSPacket {
    // {{ Class members }}

    /**
     * Flag indicating whether or not to dump extra debug info for reversing
     * the Burp internal WS packet class.
     */
    private static boolean  _show_privates = false;
    /**
     * Burp websocket message data encapsulation class
     */
    private static Class	_wsclazz = null;
    /**
     * Burp byte[] encapsulation class
     */
    private static Class	_barclazz = null;
    /**
     * Burp URL encapsulation class
     */
    private static Class	_earlclazz = null;
    /**
     * Accessor method on _barclazz instances
     */
    private static Method	_burp2bar = null;
    /**
     * Accessor methods on _wsclazz instances
     */
    private static Method	_get_comment = null;
    private static Method	_get_time = null;
    private static Method	_get_c2s = null;
    private static Method	_get_url = null;
    private static Method	_get_orgmsg = null;
    private static Method	_get_automodmsg = null;
    private static Method	_get_cgdmsg = null;
    
    // {{ Instance members }}
    
    /**
     * Encapsulated Burp websocket message object
     */
    private Object _burp_packet = null;

    // {{ Class methods }}

    /**
     * Helper method for printing messages.
     * <p>
     * The printed message is written to <code>System.out</code> and starts with the current
     * class name, followed by two dashes, followed by the specified text.
     * @param s contains the <code>String</code> value to print.
     */
    private static void say(String s) {
	System.out.println("WSPacket -- " + s);
    }

    /**
     * The Burp specific class that encapsulates a websocket message is burp.v1h and
     * it has one constructor with the following argument list:
     *
     *     public v1h(	final int k,	// local listener port
     *			final qpd c,	// URL encapsuclation
     *			final jme j,	// original message data encapsulation
     *			final jme i,	// auto modified message data encapsulation
     *			final boolean f,	// direction
     *			final wog d,	// enumaration for message data type
     *			final long h)	// timestamp
     *
     */
    private static void getByteArrayEncapsulation(Class clz) throws Exception {
	Constructor[] inits = clz.getDeclaredConstructors();
	if(inits.length != 1) {
	    throw new Exception("class " + clz.getName() +
				" expected exactly one constructor, found " +
				inits.length);
	}
	Class[] paramtypes = inits[0].getParameterTypes();
	_earlclazz = paramtypes[1];
	_barclazz = paramtypes[2];
    }

    /**
     * Analyze properties of websocket message data encapsulation class and search
     * for an accessor method that returns a byte array: that method will be used
     * for data extraction while dumping the data.
     */
    private static Method getBarExtraction(Class c) {
	Method m = null;
	Method[] mets = c.getDeclaredMethods();
	for(int i = 0; i < mets.length; ++i) {
	    Class[] paramtypes = mets[i].getParameterTypes();
	    if(paramtypes.length == 0) {
		if(mets[i].getReturnType().equals(byte[].class)) {
		    m = mets[i];
		    break;
		}
	    }
	}
	return m;
    }
    /**
     * Get accessor methods needed to process emcapsulated websocket messages.
     *
     * Code based on analysis of the actual collection Burp class. Note that the
     * obfuscator produces multiple methods with the same Java signature, but with
     * different return type. The only way to get the methods that we need is to
     * get analyze all methods declared for the class and check both argument list
     * and return value.
     *
     * Part of the methods could be derived automatically but for simplicity sake
     * we just use the information provided by ObjectDumper for all the interesting
     * accessor methods.
     *
     * @param c class encapsulating one intercepted websocket message
     */
    private static void getAccessorMethods(Class c) {
	Method[] mets = c.getDeclaredMethods();
	for(int i = 0; i < mets.length; ++i) {
	    Class[] paramtypes = mets[i].getParameterTypes();
	    if(paramtypes.length == 0) {
		if(mets[i].getReturnType().equals(long.class)) {
		    //
		    // long is typically used for timestamps and there is only one
		    // long property
		    //
		    _get_time = mets[i];
		} else if(mets[i].getReturnType().equals(boolean.class)) {
		    //
		    // boolean is typically used for items that can have only two
		    // values and there is only one boolean property
		    //
		    _get_c2s = mets[i];
		} else if(mets[i].getReturnType().equals(_barclazz) &&
			  mets[i].getName().equals("f")) {
		    _get_orgmsg = mets[i];
		} else if(mets[i].getReturnType().equals(String.class) &&
			  mets[i].getName().equals("a")) {
		    _get_comment = mets[i];
		} else if(mets[i].getReturnType().equals(_barclazz) &&
			  mets[i].getName().equals("e")) {
		    _get_automodmsg = mets[i];
		} else if(mets[i].getReturnType().equals(_barclazz) &&
			  mets[i].getName().equals("d")) {
		    _get_cgdmsg = mets[i];
		} else if(mets[i].getReturnType().equals(_earlclazz)) {
		    //
		    // Could to with further refinement for getting actual URL
		    // string: for time being just call "toString()"
		    // on whatever that is returned by this method
		    //
		    _get_url = mets[i];
		}
	    }
	}
    }

    //
    // Toggle flag that will dump additional info useful for reversing
    // the Burp internal representation of a WS packet
    //
    public static void setDebug(boolean flag) {
	_show_privates = flag;
    }
    public static boolean getDebug() {
	return _show_privates;
    }
    /**
     * Get remaining info from Burp environment for working with websocket messages.
     *
     * The code in this method and the methods it calls are using a mix of automated
     * analysis (e.g. search for the single accessor method that returns a byte array)
     * and hardcoded names obtained from dumping actual object hierarchy info at run-
     * time.
     *
     * After executing this method we have all the info we need to dump the websocket
     * messages.
     *
     * @param model TableModel for JTable showing summary info for each websocket
     * message. In BurpSuite 1.7.15 this is an instance of class burp.e8g
     *
     */
    public static Object init(Object model) {
	try {
	    //
	    // ObjectDumper output shows that property named "D" contains the list
	    // of websocket messages
	    //
	    Class<?> clz = model.getClass();	// assume never null
	    Object lst;
	    Field fld = clz.getDeclaredField("D");
	    fld.setAccessible(true);	// Make sure we can access the field
	    lst = fld.get(model);	// is an instance of burp.e0h
	    if(!(lst instanceof Collection)) {
		throw new Exception("TableModel.collection property is not a Collection but is a " +
				    lst.getClass().getName() + " instead");
	    }
	    Collection col = (Collection) lst;
	    if(col.size() <= 0) {
		throw new Exception("Websocket messages table is empty!");
	    }
	    Object[] elms = col.toArray();
	    _wsclazz = elms[0].getClass();	// = burp.v1h
	    if(getDebug()) {
		say("init() -- _wsclazz is: " + _wsclazz.getName());
		Foefel.dumpStatics(_wsclazz);
	    }
	    getByteArrayEncapsulation(_wsclazz);
	    getAccessorMethods(_wsclazz);
	    _burp2bar = getBarExtraction(_barclazz);
	    return lst;
	} catch(Throwable t) {
	    System.out.println("## caught: " + t);
	    t.printStackTrace();
	    System.exit(1);
	}
	return null;
    }

    // {{ Constructors }}
    
    /**
     * Default public constructor.
     */
    public WSPacket(Object bp) {
	_burp_packet = bp;
    }
    
    // {{ Instance methods }}

    /**
     * Accessor methods to Burp websocket message objects
     */
    public boolean getClientToServer() {
	try {
	    return ((Boolean)_get_c2s.invoke(_burp_packet)).booleanValue();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return false;
    }
    public long getTimestamp() {
	try {
	    return ((Long)_get_time.invoke(_burp_packet)).longValue();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return 0L;
    }
    public String getComment() {
	try {
	    return (String)_get_comment.invoke(_burp_packet);
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public String getURL() {
	try {
	    Object o = _get_url.invoke(_burp_packet);;
	    return _get_url.invoke(_burp_packet).toString();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public byte[] getOriginalMessage() {
	try {
	    Object o = _get_orgmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public byte[] getEditedMessage() {
	try {
	    Object o = _get_cgdmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public byte[] getAutoModifiedMessage() {
	try {
	    Object o = _get_automodmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
}
